'''
pdfsubject.py: write Zotero select link into the PDF file's subject property

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from   bun import inform, warn
from   commonpy.string_utils import antiformat
from   pdfrw import PdfReader, PdfWriter
import re

from .base import WriterMethod

if __debug__:
    from sidetrack import log


# Class definitions.
# .............................................................................

class PDFSubject(WriterMethod):
    '''Implements writing Zotero links into the PDF file's Producer property.'''

    @classmethod
    def name(cls):
        return 'pdfsubject'


    @classmethod
    def description(cls):
        return ('Writes the Zotero select link into the "Subject" metadata'
                + ' field of each PDF file. If the "Subject" field is not empty'
                + ' on a given file, Zowie looks for an existing Zotero link'
                + ' within the value and updates the link if one is found;'
                + ' otherwise, Zowie leaves the field untouched unless given'
                + ' the overwrite flag (-o), in which case, it replaces the'
                + ' entire contents of the field with the Zotero select link.'
                + ' Note that the PDF "Subject" field is not the same as the'
                + ' "Title" field. For some users, the "Subject" field is not'
                + ' used for any purpose and thus can be usefully hijacked for'
                + ' storing the Zotero select link. The value is accessible'
                + ' from macOS Preview, Adobe Acrobat, DEVONthink, and'
                + ' presumably any other application that can display the PDF'
                + ' metadata fields.')


    def write_link(self, file_path, uri):
        '''Write the "uri" into the Subject attribute of PDF file "file_path".'''

        fp = antiformat(file_path)
        if __debug__: log(f'reading PDF file {fp}')
        trailer = PdfReader(file_path)
        file = antiformat(f'[steel_blue3]{file_path}[/]')
        if not self.overwrite:
            subject = trailer.Info.Subject or ''
            if __debug__: log(f'found PDF Subject value {subject} on {fp}')
            if uri in subject:
                inform(f'Zotero link already present in PDF "Subject" field of {file}')
                return
            elif subject.startswith('zotero://select'):
                inform(f'Replacing existing Zotero link in PDF "Subject" field of {file}')
                subject = re.sub(r'(zotero://\S+)', uri, subject)
                trailer.Info.Subject = subject
            elif subject is not None:
                warn(f'Not overwriting existing PDF "Subject" value in {file}')
                return
            else:
                if __debug__: log(f'no prior PDF Subject field found on {fp}')
                inform(f'Writing Zotero link into PDF "Subject" field of {file}')
                trailer.Info.Subject = uri
        else:
            inform(f'Overwriting PDF "Subject" field of {file}')
            trailer.Info.Subject = uri

        if not self.dry_run:
            if __debug__: log(f'writing PDF file with new "Subject" field: {fp}')
            PdfWriter(file_path, trailer = trailer).write()



'''
exceptions.py: exceptions defined by Zowie

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''


# Base class.
# .............................................................................
# The base class makes it possible to use a single test to distinguish between
# exceptions generated by Zowie and exceptions generated by something else.

class ZowieException(Exception):
    '''Base class for Zowie exceptions.'''
    pass


# Exception classes.
# .............................................................................

class CannotProceed(ZowieException):
    '''A recognizable condition caused an early exit from the program.'''
    pass

class UserCancelled(ZowieException):
    '''The user elected to cancel/quit the program.'''
    pass

class NetworkFailure(ZowieException):
    '''Unrecoverable problem involving network operations.'''
    pass

class ServiceFailure(ZowieException):
    '''Unrecoverable problem involving a remote service.'''
    pass

class AuthenticationFailure(ZowieException):
    '''Problem obtaining or using authentication credentials.'''
    pass

class InternalError(ZowieException):
    '''Unrecoverable problem involving Zowie itself.'''
    pass

class NotFound(ZowieException):
    '''Item could not be found.'''
    pass

class FileError(ZowieException):
    '''Problem reading or writing a file or its attributes.'''
    pass



'''
zotero_utils.py: utilities for interacting with Zotero libraries

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from bun import inform, warn, alert, alert_fatal
from commonpy.interrupt import raise_for_interrupts
from commonpy.string_utils import antiformat
from collections import namedtuple
from os import path as path
from pyzotero import zotero, zotero_errors

from .exceptions import CannotProceed
from .exit_codes import ExitCode
from .keyring_utils import keyring_credentials, save_keyring_credentials
from .keyring_utils import validated_input

if __debug__:
    from sidetrack import log, logr


# Data definitions.
# .............................................................................

ZoteroRecord = namedtuple('ZoteroRecord', 'key parent_key file link record')
ZoteroRecord.__doc__ = '''Zotero data about a local file
  'key' is the Zotero key for the file attachment
  'parent_key' is the top-level record that contains the file attachment
  'file' is the path to the file on the local file system
  'link' is a Zotero select link of the form "zotero://select/..."
  'record' is the entire record from Zotero
'''


# Exported classes.
# .............................................................................

# maybe this should be a subclass of the pytoztero class?

class Zotero():
    '''Zotero interface class.'''

    def __init__(self, key, user_id, use_keyring):
        if key and (not key.isalnum() or len(key) < 20):
            alert_fatal(f'"{key}" does not appear to be a valid API key.')
            raise CannotProceed(ExitCode.bad_arg)
        if user_id and not user_id.isdigit():
            alert_fatal(f'"{user_id}" does not appear to be a Zotero user ID.')
            raise CannotProceed(ExitCode.bad_arg)

        # If the user supplied all the values on the command line, those are
        # the values used.  If none were supplied by the user on the command
        # line, all the values are retrieved from the user's keyring.  If
        # some were supplied and others missing, they're filled in from
        # either the keyring or by prompting the user.

        if __debug__: log('keyring {}', 'enabled' if use_keyring else 'disabled')
        if key is None and user_id is None and use_keyring:
            # We weren't given a key and id, but we can look in the keyring.
            if __debug__: log(f'getting id & key from keyring')
            key, user_id = keyring_credentials()
        if not key:
            key = validated_input('API key', key, lambda x: x.isalnum())
        if not user_id:
            user_id = validated_input('User ID', user_id, lambda x: x.isdigit())
        if use_keyring:
            if __debug__: log('saving credentials to keyring')
            save_keyring_credentials(key, user_id)

        self._key = key
        self._user_id = user_id

        # Get connected and store the Zotero conection object for the user
        # library first, then look up the group libraries that the user can
        # access and create Zotero objects for that too.  The way we use them,
        # we don't need to separate between them, so they all go into one list.

        self._libraries = []
        try:
            if __debug__: log(f'connecting to Zotero as user {user_id}')
            user = zotero.Zotero(user_id, 'user', key)
            # pyzotero will return an object but that doesn't mean the user
            # actually gave valid credentials. Need to try an operation.
            user.count_items()
            self._libraries.append(user)
            raise_for_interrupts()
        except zotero_errors.UserNotAuthorised as ex:
            if __debug__: log(f'got exception {str(ex)}')
            alert_fatal('Unable to connect to Zotero: invalid ID and/or API key.',
                        'The Zotero servers rejected attempts to connect.')
            raise CannotProceed(ExitCode.bad_arg)
        except KeyboardInterrupt as ex:
            if __debug__: log(f'got exception {str(ex)}')
            raise
        except Exception as ex:
            if __debug__: log(f'failed to create Zotero user object: str(ex)')
            alert_fatal('Unable to connect to Zotero API.')
            raise

        try:
            for group in user.groups():
                if __debug__: log(f'user can access group id {group["id"]}')
                self._libraries.append(zotero.Zotero(group['id'], 'group', key))
                raise_for_interrupts()
        except KeyboardInterrupt as ex:
            if __debug__: log(f'got exception {str(ex)}')
            raise
        except Exception as ex:
            if __debug__: log(f'failed to create Zotero group object: str(ex)')
            alert('Unable to retrieve Zotero group library; proceeding anyway.')


    def record_for_file(self, file):
        '''Returns a ZoteroRecord corresponding to the given local PDF file.'''
        f = antiformat(file)
        if not path.exists(file):
            # The file should always exist, because of how the list of files is
            # gathered, so something is wrong but we don't know what. Give up.
            raise ValueError(f'File not found: {f}')

        # Zotero stores content in the subdirectory like .../storage/N743ZXDF.
        # The item key is the alphanumeric directory name.
        # Given the key, there's no way to know whether the record is in a user
        # library or a group library, so we have to iterate over the options.
        itemkey = path.basename(path.dirname(file))
        record = None
        for library in self._libraries:
            try:
                record = library.item(itemkey)
                if __debug__: log(f'{itemkey} found in library {library.library_id}')
                break
            except zotero_errors.ResourceNotFound:
                if __debug__: log(f'{itemkey} not found in library {library.library_id}')
                continue
            except KeyboardInterrupt as ex:
                if __debug__: log(f'interrupted: {str(ex)}')
                raise
            except Exception as ex:
                if __debug__: log(f'got exception {str(ex)}')
                raise
            # pyzotero calls urllib3 for network connections. The latter uses
            # try-except clauses that broadly catch all Exceptions but don't
            # check for for KeyboardInterrupt. Thus, ^C during a network call
            # will show up as a failure to return data, not a KeyboardInterrupt.
            raise_for_interrupts()
        if not record:
            if __debug__: log(f'could not find a record for item key "{itemkey}"')
            return (None, f'Unable to retrieve Zotero record for {f}')

        # If the PDF isn't associated with a bib record, it won't have a parent.
        parentkey = self.parent_key(record, file)
        if not parentkey:
            if __debug__: log(f'file not associated with a parent record: {f}')
            return (None, f'File lacks a parent Zotero record: {f}')

        # We have an item record and a parent. We are happy campers.
        if __debug__: log(f'{parentkey} is parent of {itemkey} for {f}')
        r = ZoteroRecord(key = itemkey, parent_key = parentkey, file = file,
                         link = self.item_link(record, file), record = record)
        return (r, None)


    # For the format of the URIs, see these discussions in the Zotero forums:
    # https://forums.zotero.org/discussion/comment/335046/#Comment_335046
    # https://forums.zotero.org/discussion/78312/zotero-uri-vs-select-item

    def item_link(self, record, file):
        '''Given a record, returns an item link (i.e., "zotero://select/...)'''
        libtype = record['library']['type']
        parentkey = self.parent_key(record, file)
        if not parentkey:
            return None
        if libtype == 'user':
            return f'zotero://select/library/items/{parentkey}'
        else:
            groupid = record['library']['id']
            return f'zotero://select/groups/{groupid}/items/{parentkey}'


    def parent_key(self, record, file):
        '''Safely returns the parent key of the record item, or None.'''
        f = antiformat(file)
        if 'data' not in record:
            if __debug__: log(f'no "data" in record for {f}')
            return None
        if 'parentItem' not in record['data']:
            if __debug__: logr(f'unexpected record for {f}: ' + str(record["data"]))
            return None
        return record['data']['parentItem']



'''
keyring_utils.py: utilities for reading and writing credentials to the keyring

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from   bun import alert
import getpass
import keyring
import sys

if sys.platform.startswith('win'):
    import keyring.backends
    from keyring.backends.Windows import WinVaultKeyring
if sys.platform.startswith('darwin'):
    import keyring.backends
    from keyring.backends.OS_X import Keyring

if __debug__:
    from sidetrack import log


# Global constants.
# .............................................................................

_KEYRING = f'mhucka.{__package__}'
'''The name of the keyring used to store server access credentials, if any.'''


# Main functions.
# .............................................................................

# Explanation about the weird way this is done: the Python keyring module
# only offers a single function for setting a value; ostensibly, this is
# intended to store a password associated with an identifier (a user name),
# and this identifier is expected to be obtained some other way, such as by
# using the current user's computer login name.  But, in our situation, we
# have multiple pieces of information we have to store (a user id and an api
# key).  The hackacious solution taken here is to concatenate the values into
# a single string used as the actual value stored.  The individual values are
# separated by a character that is unlikely to be part of any user-typed value.

def keyring_credentials(ring = _KEYRING):
    '''Looks up the user's credentials.'''
    if sys.platform.startswith('win'):
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        keyring.set_keyring(Keyring())
    value = keyring.get_password(ring, getpass.getuser())
    if __debug__: log(f'got "{value}" from keyring {_KEYRING}')
    return _decoded(value) if value else (None, None)


def save_keyring_credentials(api_key, user_id, ring = _KEYRING):
    '''Saves the user's credentials.'''
    if sys.platform.startswith('win'):
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        keyring.set_keyring(Keyring())
    value = _encoded(api_key, user_id)
    if __debug__: log(f'storing "{value}" to keyring {_KEYRING}')
    keyring.set_password(ring, getpass.getuser(), value)


# Utility functions.
# .............................................................................

_SEP = ''
'''Character used to separate multiple actual values stored as a single
encoded value string.  This character is deliberately chosen to be something
very unlikely to be part of a legitimate string value typed by user at a
shell prompt, because control-c is normally used to interrupt programs.
'''

def _encoded(api_key, user_id):
    return f'{api_key}{_SEP}{user_id}'


def _decoded(value_string):
    return tuple(value_string.split(_SEP))


def password(prompt):
    # If it's a tty, use the version that doesn't echo the password.
    if sys.stdin.isatty():
        return getpass.getpass(prompt)
    else:
        sys.stdout.write(prompt)
        sys.stdout.flush()
        return sys.stdin.readline().rstrip()


def validated_input(msg, default_value, is_valid):
    while True:
        if __debug__: log(f'asking user: "{msg} [{default_value}]"')
        default = (' [' + default_value + ']') if default_value else ''
        value = input(msg + default + ': ')
        if default_value and value == '':
            if __debug__: log(f'user chose default value "{default_value}"')
            return default_value
        elif is_valid(value):
            if __debug__: log(f'got "{value}" from user')
            return value
        else:
            alert(f'"{value}" does not appear valid for {msg}')
            return None



'''
wherefrom.py: writes Zotero select link into the "Where from" metadata field

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

import applescript
import biplist
from   bun import inform, warn
from   commonpy.string_utils import antiformat
import re
from   xattr import getxattr, setxattr, listxattr

from .base import WriterMethod
from ..exceptions import FileError

if __debug__:
    from sidetrack import log


# Class definitions.
# .............................................................................

class WhereFrom(WriterMethod):
    '''Implements writing Zotero links into the "Where from" metadata field.'''

    @classmethod
    def name(cls):
        return 'wherefrom'


    @classmethod
    def description(cls):
        return ('Writes the Zotero select link to the "Where from" metadata'
                + ' field of each file (the com.apple.metadata:kMDItemWhereFroms'
                + ' extended attribute). This field is displayed as "Where from"'
                + ' in Finder "Get Info" panels; it is typically used by web'
                + ' browsers to store a files download origin. The field is a'
                + ' list. If Zowie finds a Zotero select link as the first item'
                + ' in the list, it updates its value; otherwise, Zowie prepends'
                + ' the Zotero select link to the list of existing values,'
                + ' keeping the other values unless the overwrite option (-o) is'
                + ' used. When the overwrite option is used, Zowie deletes the'
                + ' existing list of values and writes only the Zotero select'
                + ' link. Note that if macOS Spotlight indexing is turned on for'
                + ' the volume containing the file, the macOS Finder will'
                + ' display the upated "Where from" values in the Get Info panel'
                + ' of the file; if Spotlight is not turned on, the Get info'
                + ' panel will not be updated, but other applications will'
                + ' still be able to read the updated value.')


    def write_link(self, file_path, uri):
        '''Write the "uri" into the "Where From" metadata attribute of "file_path".'''

        # file pathname string may contain '{' and '}', so guard against it.
        fp = antiformat(file_path)
        file = antiformat(f'[steel_blue3]{file_path}[/]')
        if not self.overwrite:
            (wherefroms, malformed) = self._wherefroms(file_path)
            if wherefroms:
                if wherefroms[0] == uri:
                    inform(f'Zotero link already present in "Where from" of {file}')
                    # We found a link already present, but the attribute value
                    # was malformed (maybe due to the use of a buggy previous
                    # version of Zowie or manual experiments by the user). We
                    # should proceed to correct it even if -o is not in effect.
                    if not malformed:
                        return
                elif type(wherefroms[0]) is str and wherefroms[0].startswith('zotero://'):
                    inform(f'Updating existing Zotero link in "Where from" of {file}')
                    wherefroms[0] = uri
                else:
                    inform(f'Prepending Zotero link to front of "Where from" of {file}')
                    wherefroms.insert(0, uri)
            else:
                if __debug__: log(f'no prior wherefroms found on {fp}')
                inform(f'Writing Zotero link into "Where From" metadata of {file}')
                wherefroms = [uri]
        else:
            inform(f'Overwriting "Where From" metadata with Zotero link in {file}')
            wherefroms = [uri]

        self._write_wherefroms(file_path, wherefroms)


    def _wherefroms(self, file_path):
        '''Returns a tuple (wherefroms, malformed), where the second element
        indicates where the content was malformed in some way.'''

        fp = antiformat(file_path)
        if b'com.apple.metadata:kMDItemWhereFroms' in listxattr(file_path):
            wherefroms = getxattr(file_path, b'com.apple.metadata:kMDItemWhereFroms')
            if not wherefroms.startswith(b'bplist'):
                # There's content, but it's not a list. We don't know how to
                # parse it and can't anticipate every possible variation, but
                # we shouldn't leave it or destroy it either, if we can help it.
                if __debug__: log(f'wherefroms {wherefroms} is not a list on {fp}')
                try:
                    # We want to return an array of strings. Here, we have
                    # bytes. Try to convert it, just in case we succeed.
                    wherefroms = wherefroms.decode()
                except UnicodeDecodeError:
                    raise FileError('Malformed non-list value for attribute'
                                    + ' com.apple.metadata:kMDItemWhereFroms'
                                    + f' on file {fp}')
                return ([wherefroms], True)
            try:
                wherefroms = biplist.readPlistFromString(wherefroms)
            except biplist.InvalidPlistException as ex:
                if __debug__: log(f'got exception {str(ex)} parsing wherefroms of {fp}')
                # The property exists, it looks like a list, but it failed to
                # parse. We can't treat it as preexisting content because it'll
                # probably screw up something else. If we're going to overwrite
                # it anyway, we won't get to this point anyway. Otherwise:
                raise FileError('Unable to parse attribute'
                                + ' com.apple.metadata:kMDItemWhereFroms'
                                + f' on file {fp}')
            return (wherefroms, False)
        else:
            if __debug__: log(f'no kMDItemWhereFroms attribute on {fp}')
            return (None, False)


    def _write_wherefroms(self, file_path, wherefroms):
        binary = biplist.writePlistToString(wherefroms)
        if __debug__: log(f'writing "where froms" attribute of {antiformat(file_path)}')
        setxattr(file_path, b'com.apple.metadata:kMDItemWhereFroms', binary)



'''
findercomment.py: write Zotero select link into a file's macOS Finder comments

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

import applescript
from   bun import inform, warn
from   commonpy.string_utils import antiformat
import re

from .base import WriterMethod

if __debug__:
    from sidetrack import log


# Constants.
# .............................................................................

# The following code is based in part on code from Python package "osxmetadata"
# version 0.99.10 of 2020-09-01, Copyright (c) 2020 Rhet Turnbull.
# https://github.com/RhetTbull/osxmetadata/blob/master/osxmetadata/findercomments.py

_FINDER_SCRIPTS = applescript.AppleScript('''
on get_comments{f}
    tell application "Finder"
        return comment of (POSIX file f as alias)
    end tell
end run

on set_comments{f, c}
    tell application "Finder"
        set comment of (POSIX file f as alias) to c as Unicode text
    end tell
end run

on clear_comments{f}
    tell application "Finder"
        set c to missing value
        set comment of (POSIX file f as alias) to c
    end tell
end run
''')


# Class definitions.
# .............................................................................

class FinderComment(WriterMethod):
    '''Implements writing Zotero URI into the macOS Finder comments for a file.'''

    @classmethod
    def name(cls):
        return 'findercomment'


    @classmethod
    def description(cls):
        return ('(Default method.)'
                + ' Writes the Zotero select link into the Finder comments of'
                + ' each file, attempting to preserve other parts of the'
                + ' comments. If Zowie finds an existing Zotero select link in'
                + ' the text of the Finder comments attribute, it only updates'
                + ' the link portion and tries to leave the rest of the'
                + ' comment text untouched. Otherwise, Zowie ONLY writes'
                + ' into the comments attribute if either the attribute value is'
                + ' empty or Zowie is given the overwrite (-o) option. (Note'
                + ' that updating the link text requires rewriting the entire'
                + ' Finder comments attribute on a given file. Finder comments'
                + ' have a reputation for being easy to get into inconsistent'
                + ' states, so if you have existing Finder comments that you'
                + " absolutely don't want to lose, it may be safest to avoid"
                + ' this method.)')


    def write_link(self, file_path, uri):
        '''Writes the "uri" into the Finder comments of file "file_path".

        If there's an existing comment, read it.  If there's a Zotero select
        link as the first thing in the comment, replace that URI with this one,
        under the assumption that this was a link written by a prior run of
        this program; otherwise, prepend the Zotero select link to the finder
        comments.  In either case, write the results back.
        '''

        # file pathname string may contain '{' and '}', so guard against it.
        fp = antiformat(file_path)
        file = antiformat(f'[steel_blue3]{file_path}[/]')
        if not self.overwrite:
            if __debug__: log(f'reading Finder comments of {fp}')
            comments = _FINDER_SCRIPTS.call('get_comments', file_path)
            if comments and uri in comments:
                inform(f'Zotero link already present in Finder comments of {file}')
                return
            elif comments and 'zotero://select' in comments:
                inform(f'Replacing existing Zotero link in Finder comments of {file}')
                if __debug__: log(f'overwriting existing Zotero link with {uri}')
                comments = re.sub(r'(zotero://\S+)', uri, comments)
            elif comments:
                warn(f'Not overwriting existing Finder comments of {file}')
                return
            else:
                inform(f'Writing Zotero link into empty Finder comments of {file}')
                comments = uri
        else:
            inform(f'Ovewriting Finder comments with Zotero link for {file}')
            comments = uri

        if not self.dry_run:
            if __debug__: log(f'invoking AS function to clear comment on {fp}')
            _FINDER_SCRIPTS.call('clear_comments', file_path)
            if __debug__: log(f'invoking AS function to set comment on {fp}')
            _FINDER_SCRIPTS.call('set_comments', file_path, comments)



'''
Zowie: a tool to write Zotero select links into Zotero article PDF files

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under the MIT license.  Please see
the file "LICENSE" for more information.
'''

# Package metadata ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
#  ╭────────────────────── Notice ── Notice ── Notice ─────────────────────╮
#  |    The following values are automatically updated at every release    |
#  |    by the Makefile. Manual changes to these values will be lost.      |
#  ╰────────────────────── Notice ── Notice ── Notice ─────────────────────╯

__version__     = '1.0.5'
__description__ = 'Write Zotero select links into article PDF files'
__url__         = 'https://github.com/mhucka/zowie'
__author__      = 'Michael Hucka'
__email__       = 'mhucka@caltech.edu'
__license__     = 'BSD 3-clause'


# Miscellaneous utilities.
# .............................................................................

def print_version():
    print(f'{__name__} version {__version__}')
    print(f'Authors: {__author__}')
    print(f'URL: {__url__}')
    print(f'License: {__license__}')



'''
Zowie: a program to write Zotero select links into article PDF files

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from   bun import UI, inform, warn, alert, alert_fatal
from   boltons.debugutils import pdb_on_signal
from   commonpy.data_utils import timestamp
from   commonpy.interrupt import config_interrupt
from   commonpy.string_utils import antiformat
import plac
import signal
import shutil
import sys
from   sys import exit as exit
from   textwrap import wrap

import zowie
from zowie import print_version
from .exceptions import UserCancelled, FileError, CannotProceed
from .exit_codes import ExitCode
from .main_body import MainBody
from .methods import method_names, KNOWN_METHODS

if __debug__:
    from sidetrack import set_debug, log, logr


# Main program.
# .............................................................................

@plac.annotations(
    api_key    = ('API key to access the Zotero API service',                'option', 'a'),
    no_color   = ('do not color-code terminal output',                       'flag',   'C'),
    after_date = ('only act on files created or modified after date "D"',    'option', 'd'),
    identifier = ('Zotero user ID for API calls',                            'option', 'i'),
    no_keyring = ('do not store credentials in the keyring service',         'flag',   'K'),
    list       = ('print list of known methods',                             'flag',   'l'),
    method     = ('select method to store links (default: finder comments)', 'option', 'm'),
    dry_run    = ('report what would be done without actually doing it',     'flag',   'n'),
    overwrite  = ('forcefully overwrite previous content',                   'flag',   'o'),
    quiet      = ('be less chatty -- only print important messages',         'flag',   'q'),
    version    = ('print version info and exit',                             'flag',   'V'),
    debug      = ('write detailed trace to "OUT" ("-" means console)',       'option', '@'),
    files      = 'file(s) and/or folder(s) containing Zotero article PDF files',
)

def main(api_key = 'A', no_color = False, after_date = 'D', identifier = 'I',
         no_keyring = False, list = False, method = 'M', dry_run = False,
         overwrite = False, quiet = False, version = False, debug = 'OUT', *files):
    '''Zowie ("ZOtero link WrItEr") is a tool for Zotero users.

Zowie writes Zotero select links into the PDF files and/or the macOS Finder
metadata attributes of PDF files in the user's Zotero database. This makes it
possible to look up the Zotero entry of a PDF file from outside of Zotero.

Credentials for Zotero access
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Zowie needs to know the user's personal library identifier (also known as the
"userID") and a Zotero API key. By default, it tries to get this information
from the user's keychain. If the values do not exist in the keychain from a
previous run, Zowie will ask the user, and (unless the -K option is given)
store the values in the user's keychain so that it does not have to ask again
in the future. It is also possible to supply the identifier and API key on
the command line using the -i and -a options, respectively; the given values
will then override the values stored in the keychain (unless the -K option is
also given). This is also how you can replace previously-stored values: use
-i and -a (without -K) and the new values will override the stored values.

To find out your Zotero userID and create an API key, log in to your Zotero
account at Zotero.org and visit https://www.zotero.org/settings/keys

Basic usage
~~~~~~~~~~~

Zowie can operate on a folder, or one or more individual PDF files, or a mix
of both. Suppose your local Zotero database is located in ~/Zotero/. Perhaps
the simplest way to run Zowie is the following command:

  zowie ~/Zotero

If this is your first run of Zowie, it will ask you for your userID and API
key, then search for PDF files recursively under ~/Zotero/.  For each PDF
file found, Zowie will contact the Zotero servers over the network and
determine the Zotero URI for the bibliographic entry containing that PDF
file. Finally, it will use its default method of writing the Zotero select
link, which is to write it into the macOS Finder comments for the file.

Instead of a folder, you can invoke Zowie on one or more individual files (but
be careful to quote pathnames with spaces in them, such as in this example):

  zowie "~/Zotero/storage/26GS7CZL/Smith 2020 Paper.pdf"

Zowie supports multiple methods of writing the Zotero select link.  The
option -l will cause Zowie to print a list of all the methods available:

  zowie -l

The default method is to write it into Finder comments for the file. (These
comments are visible in the Finder's "Get Info" panel.) The option -m can be
used to select one or more alternative methods. Separate the names with
commas without spaces. For example, the following command will make Zowie
write the Zotero link into the Finder comments and the "Where from" attribute:

  zowie -m findercomment,wherefrom ~/Zotero/storage

Where possible, Zowie tries to preserve the previous contents of metadata
attributes.  For example, In the case of Finder comments and "Where from", it
looks for existing Zotero links in the contents and updates those links only;
if it does not find an existing Zotero link, it prepends one instead of
replacing the value completely.  The general rule is that Zowie will try to
detect whether a Zotero select link is already present in the chosen metadata
attribute(s) and will only update the link text if a link is found;
otherwise, it will not write the Zotero select link at all unless given the
overwrite (-o) option.  The overwrite option (-o) makes Zowie replace values
completely.  Check the description of the methods for more details about what
they do by default and the impact of the -o option.

Filtering by date
~~~~~~~~~~~~~~~~~

If the -d option is given, the PDF files will be filtered to use only those
whose last-modified date/time stamp is no older than the given date/time
description. Valid descriptors are those accepted by the Python dateparser
package. Make sure to enclose descriptions within single or double
quotes. Examples:

 zowie -d "2 weeks ago" ....
 zowie -d "2014-08-29" ....
 zowie -d "12 Dec 2014" ....
 zowie -d "July 4, 2013" ....

Additional command-line arguments
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If given the -n ("dry run") option, Zowie will only print what it would do
without actually doing it.

If given the -q option, Zowie will not print its usual informational messages
while it is working. It will only print messages for warnings or errors.
By default messages printed by Zowie are also color-coded. If given the
option -C, Zowie will not color the text of messages it prints. (This latter
option is useful when running Zowie within subshells inside other environments
such as Emacs.)

If given the -V option, this program will print the version and other
information, and exit without doing anything else.

If given the -@ argument, this program will output a detailed trace of what it
is doing. The debug trace will be sent to the given destination, which can
be '-' to indicate console output, or a file path to send the output to a file.

When -@ has been given, Zowie also installs a signal handler on signal SIGUSR1
that will drop Zowie into the pdb debugger if the signal is sent to the
running process.

Return values
~~~~~~~~~~~~~

This program exits with a return code of 0 if no problems are encountered.
It returns a nonzero value otherwise. The following table lists the possible
return values:

  0 = success -- program completed normally
  1 = the user interrupted the program's execution
  2 = encountered a bad or missing value for an option
  3 = no network detected -- cannot proceed
  4 = file error -- encountered a problem with a file
  5 = server error -- encountered a problem with a server
  6 = an exception or fatal error occurred

Command-line arguments summary
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'''

    # Set up debug logging as soon as possible, if requested ------------------

    if debug != 'OUT':
        if __debug__: set_debug(True, debug)
        import faulthandler
        faulthandler.enable()
        if not sys.platform.startswith('win'):
            # Even with a different signal, I can't get this to work on Win.
            pdb_on_signal(signal.SIGUSR1)

    # Preprocess arguments and handle early exits -----------------------------

    ui = UI('Zowie', 'ZOtero link WrItEr', use_color = not no_color, be_quiet = quiet)
    ui.start()

    if version:
        print_version()
        exit(int(ExitCode.success))
    if list:
        inform('Known methods:\n')
        width = (shutil.get_terminal_size().columns - 2) or 78
        for name in method_names():
            text = f'[cyan2]{name}[/]: ' + KNOWN_METHODS[name].description()
            inform('\n'.join(wrap(text, width = width, subsequent_indent = '  ')))
            inform('')
        exit(int(ExitCode.success))

    methods_list = ['findercomment'] if method == 'M' else method.lower().split(',')
    bad_name = next((n for n in methods_list if n not in method_names()), None)
    if bad_name:
        alert(f'Unrecognized method name "{bad_name}".')
        alert('The available methods are: ' + ', '.join(method_names()) + '.')
        exit(int(ExitCode.bad_arg))

    # Do the real work --------------------------------------------------------

    if __debug__: log('='*8 + f' started {timestamp()} ' + '='*8)
    body = exception = None
    try:
        body = MainBody(files       = files,
                        api_key     = None if api_key == 'A' else api_key,
                        user_id     = None if identifier == 'I' else identifier,
                        use_keyring = not no_keyring,
                        after_date  = None if after_date == 'D' else after_date,
                        methods     = methods_list,
                        dry_run     = dry_run,
                        overwrite   = overwrite)
        config_interrupt(body.stop, UserCancelled(ExitCode.user_interrupt))
        body.run()
        exception = body.exception
    except Exception as ex:
        exception = sys.exc_info()

    # Try to deal with exceptions gracefully ----------------------------------

    exit_code = ExitCode.success
    if exception:
        if __debug__: log(f'main body raised exception: {antiformat(exception)}')
        if exception[0] == CannotProceed:
            exit_code = exception[1].args[0]
        elif exception[0] == FileError:
            alert_fatal(antiformat(exception[1]))
            exit_code = ExitCode.file_error
        elif exception[0] in [KeyboardInterrupt, UserCancelled]:
            warn('Interrupted.')
            exit_code = ExitCode.user_interrupt
        else:
            msg = antiformat(exception[1])
            alert_fatal(f'Encountered error {exception[0].__name__}: {msg}')
            exit_code = ExitCode.exception
            if __debug__:
                from traceback import format_exception
                details = ''.join(format_exception(*exception))
                logr(f'Exception: {msg}\n{details}')
    else:
        inform('Done.')

    # And exit ----------------------------------------------------------------

    if __debug__: log('_'*8 + f' stopped {timestamp()} ' + '_'*8)
    if __debug__: log(f'exiting with exit code {exit_code}')
    exit(int(exit_code))


# Main entry point.
# .............................................................................

# The following entry point definition is for the console_scripts keyword
# option to setuptools.  The entry point for console_scripts has to be a
# function that takes zero arguments.
def console_scripts_main():
    plac.call(main)


# The following allows users to invoke this using "python3 -m zowie".
if __name__ == '__main__':
    plac.call(main)



'''
Zowie module for implementing different methods of writing Zotero links

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from .findercomment import FinderComment
from .pdfsubject import PDFSubject
from .pdfproducer import PDFProducer
from .wherefrom import WhereFrom

KNOWN_METHODS = {
    'findercomment': FinderComment,
    'wherefrom': WhereFrom,
    'pdfsubject': PDFSubject,
    'pdfproducer': PDFProducer,
}

# Save this list to avoid recreating it all the time.
_METHOD_NAMES = sorted(KNOWN_METHODS.keys())

def method_names():
    return _METHOD_NAMES



'''
Zowie: a program to write Zotero select links into Zotero article PDF files

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from   bun import inform, warn, alert, alert_fatal
from   commonpy.data_utils import DATE_FORMAT, pluralized, parsed_datetime
from   commonpy.file_utils import filename_extension, files_in_directory
from   commonpy.network_utils import net, network_available
from   commonpy.string_utils import antiformat
from   datetime import datetime
from   os import path
from   pathlib import Path
import sys

from .exceptions import CannotProceed
from .exit_codes import ExitCode
from .methods import KNOWN_METHODS, method_names
from .zotero import Zotero

if __debug__:
    from sidetrack import log


# Exported classes.
# .............................................................................

class MainBody():
    '''Main body for Zowie.'''

    def __init__(self, **kwargs):
        '''Initialize internal state.'''

        # Assign parameters to self to make them available within this object.
        for key, value in kwargs.items():
            if __debug__: log(f'parameter value self.{key} = {value}')
            setattr(self, key, value)

        # We expose attribute "exception" that callers can use to find out
        # if the thread finished normally or with an exception.
        self.exception = None

        # Create and initialize objects for the URI writers we will use.
        self._writers = []
        for method_name in self.methods:
            method = KNOWN_METHODS[method_name](self.dry_run, self.overwrite)
            self._writers.append(method)


    def run(self):
        '''Run the main body.'''

        if __debug__: log('running MainBody')
        try:
            self._do_preflight()
            self._do_main_work()
        except Exception as ex:
            if __debug__: log(f'exception in main body: {str(ex)}')
            self.exception = sys.exc_info()
        if __debug__: log('finished MainBody')


    def stop(self):
        '''Stop the main body.'''
        if __debug__: log('stopping ...')
        pass


    def _do_preflight(self):
        '''Check the option values given by the user, and do other prep.'''

        if not network_available():
            alert_fatal('No network connection.')
            raise CannotProceed(ExitCode.no_network)

        # Sanity-check the arguments ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        hint = '(Hint: use -h for help.)'

        if not self.use_keyring and not any([self.api_key, self.user_id]):
            alert_fatal(f"Need Zotero credentials if not using keyring. {hint}")
            raise CannotProceed(ExitCode.bad_arg)
        if any(item.startswith('-') for item in self.files):
            bad = next(item for item in self.files if item.startswith('-'))
            alert_fatal(f'Unrecognized option "{bad}" in arguments. {hint}')
            raise CannotProceed(ExitCode.bad_arg)

        if self.after_date:
            try:
                # Convert user's input into a canonical format.
                self.after_date = parsed_datetime(self.after_date)
                self.after_date_str = self.after_date.strftime(DATE_FORMAT)
                if __debug__: log(f'parsed after_date as {self.after_date_str}')
            except KeyboardInterrupt as ex:
                if __debug__: log(f'got exception {str(ex)}')
                raise
            except Exception as ex:
                alert_fatal(f'Unable to parse after_date: "{str(ex)}". {hint}')
                raise CannotProceed(ExitCode.bad_arg)

        # Set up Zotero connection and gather files for work ~~~~~~~~~~~~~~~~~~

        self._zotero = Zotero(self.api_key, self.user_id, self.use_keyring)

        self._files = []
        if __debug__: log('gathering list of PDF files ...')
        for item in self.files:
            # When printing, need to guard against names containing '{' or '}'.
            fp = antiformat(item)
            if path.isfile(item) and filename_extension(item) == '.pdf':
                self._files.append(item)
            elif path.isdir(item):
                if __debug__: log(f'adding PDF files in subdirectory {fp}')
                self._files += files_in_directory(item, extensions = ['.pdf'])
            else:
                warn(f'Not a PDF file or folder of files: "{fp}"')
        # Remove ._ files, which are not actual content files but rather are
        # used by macOS to store extended attributes on non-HFS volumes.
        if __debug__: log('filtering out ._ files')
        self._files = [f for f in self._files if not path.basename(f).startswith('._')]
        if __debug__: log(f'gathered {pluralized("PDF file", self._files, True)}')

        if self.after_date:
            if __debug__: log(f'filtering files by date {self.after_date_str}')
            kept = []
            tzinfo = self.after_date.tzinfo
            for pdffile in self.files:
                mtime = datetime.fromtimestamp(Path(pdffile).stat().st_mtime)
                if mtime.replace(tzinfo = tzinfo) >= self.after_date:
                    if __debug__: log(f'keeping {pdffile}')
                    kept.append(pdffile)
            self._files = kept

        if not self._files:
            alert_fatal('No PDF files to process; quitting.')
            raise CannotProceed(ExitCode.bad_arg)


    def _do_main_work(self):
        if self.overwrite:
            warn('Overwrite mode in effect.')
        if self.dry_run:
            warn('Running in dry run mode – will not modify files.')
        inform(f'Will process {pluralized("PDF file", self._files, True)}'
               + f' using {pluralized("method", self.methods)}'
               + f' [cyan2]{", ".join(self.methods)}[/].')
        if len(self._files) > 10000:
            inform("(That's a huge number of files – this will take a long time.)")
        elif len(self._files) > 1000:
            inform("(That's a lot of files – this will take some time.)")

        for pdffile in self._files:
            (record, failure) = self._zotero.record_for_file(pdffile)
            if failure:
                warn(failure)
                continue
            for method in self._writers:
                method.write_link(pdffile, record.link)



'''
base.py: base class definition for annotation methods in Zowie

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2020 by Michael Hucka and the California Institute of Technology.
This code is open-source software released under a 3-clause BSD license.
Please see the file "LICENSE" for more information.
'''

from abc import ABC, abstractmethod

if __debug__:
    from sidetrack import log


# Class definitions.
# .............................................................................
# Basics for the __eq__ etc. methods came from
# https://stackoverflow.com/questions/1061283/lt-instead-of-cmp

class WriterMethod(ABC):
    '''Base class for Zotero select link writer methods.'''

    def __init__(self, dry_run = False, overwrite = False):
        self.dry_run = dry_run
        self.overwrite = overwrite


    def __str__(self):
        return self.name()


    def __repr__(self):
        return self.name()


    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not self.name() < other.name() and not other.name() < self.name()


    def __ne__(self, other):
        return not self.__eq__(self, other)


    def __lt__(self, other):
        return self.name() < other.name()


    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return other.name() < self.name()


    def __le__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not other.name() < self.name()


    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return not self.name() < other.name()


    @property
    @classmethod
    @abstractmethod
    def name(cls):
        '''Returns the canonical internal name for this method.'''
        pass


    @property
    @classmethod
    @abstractmethod
    def description(cls):
        '''Returns a description what this method does.'''
        pass


    @abstractmethod
    def write_link(self, file, uri):
        '''Write the link into the file.'''
        pass
